/*
 * tree.c - simple mutlidimensional tree
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "xmalloc.h"
#include "log.h"
#include "tree.h"

tree_node_t tree_node_new()
{
	tree_node_t node;

	node = xmalloc0(sizeof(*node));
	node->parent = NULL;
	node->next = NULL;
	node->child = NULL;

	return node;
}

void tree_node_free(tree_node_t *node)
{
	if (!node || *node == NULL)
		error("tree_node_free: node NULL");

	if ((*node)->child || (*node)->parent)
		warn("tree_node_free: node %p node->data=%p in list",
				*node, (*node)->data);
	xfree(*node);
	*node = NULL;
}

void tree_node_free_path(tree_node_t *node)
{
	if (!node || *node == NULL)
		error("tree_node_free_path: node NULL");

	tree_node_t iter, next, parent = NULL;

	/* don't use recursion here */

	iter = *node;
	while (iter) {
		if (iter->child) {
			parent = iter;
			iter = iter->child;
		} else {
			next = iter->next;
			xfree(iter);
			if (next == NULL && parent) {
				parent->child = NULL;
				iter = parent;
				parent = iter->parent;
			} else
				iter = next;
		}
	}

	*node = NULL;
}

/*
 * tree_node_add
 *
 * where: 1 - add to the beginning of the list
 *        0 - add to the end of the list
 */
void tree_node_add(tree_node_t parent, tree_node_t node, int where)
{
	tree_node_t iter;

	if (!node || !parent)
		error("tree_node_add: node or parent NULL");

	if (where) {
		if (parent->child != NULL)
			node->next = parent->child;
		else
			node->next = NULL;
		parent->child = node;
		node->parent = parent;
	} else {
		/* traverse the list to last child */
		for (iter = parent->child; iter && iter->next != NULL;
			iter = iter->next);

		if (iter != NULL)
			iter->next = node;
		else
			parent->child = node;
		node->next = NULL;
		node->parent = parent;
	}
}

void tree_node_reverse(tree_node_t root)
{
	tree_node_t iter, reverse = NULL;

	if (!root)
		error("tree_node_reverse: root NULL");

	while (root->child) {
		iter = root->child;
		root->child = iter->next;
		iter->next = reverse;
		reverse = iter;
	}

	root->child = reverse;
}

/*
 * tree_node_add_sibling
 *
 * where: 0 - add node as the right sibling
 * 	  1 - add node as the left sibling
 */
void tree_node_add_sibling(tree_node_t sibling, tree_node_t node, int where)
{
	tree_node_t iter, prev = NULL;

	if (!sibling || !node)
		error("tree_node_add_sibling: sibling or node NULL");

	if (sibling->parent == NULL)
		error("tree_node_add_sibling: sibling is root!");

	if (where) {
		/* iterate until 'sibling' is found */
		for (iter = sibling->parent->child;
		     iter != NULL; iter = iter->next) {
			if (iter == sibling)
				break;
			prev = iter;
		}
		if (prev)
			prev->next = node;
		else
			sibling->parent->child = node;
		node->next = sibling;
		node->parent = sibling->parent;
	} else {
		if (sibling->next)
			node->next = sibling->next;
		else
			node->next = NULL;
		sibling->next = node;
		node->parent = sibling->parent;
	}
}

tree_node_t tree_node_find(tree_node_t root, void *data, int child_nodes,
		int (*compare)(const void *node_data, const void *data))
{
	tree_node_t iter;
	tree_node_t upper = NULL;

	if (!root)
		error("tree_node_find: root NULL");

	if (child_nodes) {
		iter = root;
		while (iter) {
			if (compare) {
				if (compare(iter->data, data) == 0)
					return iter;
			} else {
				if (iter->data == data)
					return iter;
			}
			if (iter->child) {
				upper = iter;
				iter = iter->child;
			} else {
				if (iter->next == NULL && upper) {
					while (upper) {
						if (upper->next)
							break;
						else
							upper = upper->parent;
					}
					if (upper) {
						iter = upper->next;
						upper = iter->parent;
					} else
						iter = NULL;
				} else if (iter->next == NULL && !upper)
					iter = NULL;
				else
					iter = iter->next;
			}
		}
	} else {
		iter = root;
		while (iter) {
			if (compare) {
				if (compare(iter->data, data) == 0)
					return iter;
			} else {
				if (iter->data == data)
					return iter;
			}
			iter = iter->next;
		}
	}
	return NULL;
}

tree_node_t tree_node_find_child(tree_node_t parent, void *data,
		int (*compare)(const void *node_data, const void *data))
{
	tree_node_t iter;

	if (!parent)
		error("tree_node_find_child: parent NULL");

	for (iter = parent->child; iter != NULL; iter = iter->next) {
		if (compare) {
			if ((*compare)(iter->data, data) == 0)
				return iter;
		} else {
			if (iter->data == data)
				return iter;
		}
	}
	return NULL;
}

tree_node_t tree_node_del(tree_node_t node)
{
	tree_node_t iter, prev = NULL;

	if (!node || !node->parent)
		error("tree_node_del: node is root!");

	for (iter = node->parent->child; iter != NULL; iter = iter->next) {
		if (iter == node) {
			if (prev)
				prev->next = iter->next;
			else
				node->parent->child = iter->next;
			node->parent = NULL;
			node->next = NULL;
			return iter;
		}
		prev = iter;
	}

	return NULL;
}

void tree_node_traverse(tree_node_t root, treecallback_t callback,
		int child_nodes, void *opaque)
{
	tree_node_t iter;
	tree_node_t upper = NULL;

	if (!root)
		error("tree_node_traverse: root NULL");

	if (!callback)
		error("tree_node_traverse: callback NULL");

	/* First call callback on the parent then on children.
	 * This behaviour is more suitable because we want to
	 * know something about the parent before it's ancestors.
	 */
	if (child_nodes) {
		iter = root;
		while (iter) {

			if (callback(iter, opaque))
				return;

			if (iter->child) {
				upper = iter;
				iter = iter->child;
			} else {
				if (iter->next == NULL && upper) {
					while (upper) {
						if (upper->next)
							break;
						else
							upper = upper->parent;
					}
					if (upper) {
						iter = upper->next;
						upper = iter->parent;
					} else
						iter = NULL;
				} else if (iter->next == NULL && !upper)
					iter = NULL;
				else
					iter = iter->next;
			}
		}
	} else {
		iter = root;
		while (iter) {
			if (callback(iter, opaque))
				return;
			iter = iter->next;
		}
	}
}
